import asyncio

from aio_binance.futures.usdt import Client, WsClient
from loguru import logger


class UserSession:

    def __init__(self, key: str, secret: str, calback_event):
        self.calback_event = calback_event
        self._api = Client(key, secret)

    async def __aenter__(self):
        res = await self._api.new_listen_key()
        logger.log("ACCOUNT", '  listenKey add success  Connected')
        await asyncio.wait(
            [
                WsClient(res['listenKey']).user_data(self.calback_event),
                self.__update_key()
            ],
            return_when=asyncio.FIRST_COMPLETED
        )
        raise asyncio.CancelledError

    async def __aexit__(self, exc_type, exc, tb):
        if exc_type is not None:
            logger.error(exc)
        logger.log("ACCOUNT", '  Close User Session')
        await self._api.close_listen_key()

    async def __update_key(self):
        """
        A User Data Stream is valid for 60 minutes after api.new_listen_key()
        Doing a on a will extend its validity for 60 minutes.

        Args:
            api (Client): instance Api client
            sleep (int, optional): [description]. Defaults to 55 min.
        """
        sleep = 60*55
        first = True
        while True:
            if first:
                first = False
                await asyncio.sleep(sleep)
            await self._api.update_listen_key()
            logger.log("ACCOUNT", f'  Success update listen Key. Try 55 min...')
            await asyncio.sleep(sleep)



from sys import stderr

from loguru import logger

from aio_binance.__version__ import __version__
from .api.query import Api
from .websocket.query import Ws

logger.remove()
logger.add(stderr, colorize=True,
           format="<green>{level}</green>:     <cyan>{message}</cyan>",
           enqueue=True)


class Client(Api):

    logger.level("API", no=10)

    def __init__(
            self,
            key: str = None,
            secret: str = None,
            testnet: bool = False,
            debug: str = "error"):
        logger.level(debug.upper())
        super().__init__(key=key, secret=secret, testnet=testnet, version=__version__)

    # ACCOUNT(including orders and trades)
    from .api.account import (account, adl_quantile, api_trading_status,
                              balance, cancel_batch_order, cancel_open_orders,
                              cancel_order, change_leverage,
                              change_margin_type, change_multi_asset_mode,
                              change_position_mode, commission_rate,
                              countdown_cancel_order, force_orders,
                              get_account_trades, get_all_orders,
                              get_income_history, get_multi_asset_mode,
                              get_open_orders, get_orders,
                              get_position_margin_history, get_position_mode,
                              get_position_risk, leverage_brackets,
                              modify_isolated_position_margin, new_batch_order,
                              new_order, new_order_test, query_order)
    # MARKETS
    from .api.market import (agg_trades, asset_Index, blvt_kline, book_ticker,
                             continuous_klines, depth, exchange_info,
                             funding_rate, historical_trades, index_info,
                             index_price_klines, klines, mark_price,
                             mark_price_klines, open_interest, ping,
                             ticker_24hr_price_change, ticker_price, time,
                             trades)
    # STREAMS
    from .api.stream import close_listen_key, new_listen_key, update_listen_key


class WsClient(Ws):
    logger.level("ACCOUNT", no=10)
    logger.level("WEBSOCKET", no=10)

    def __init__(
            self,
            listen_key: str = None,
            debug='error',
            **kwargs):
        logger.level(debug.upper())
        super().__init__(listen_key=listen_key, **kwargs)

    from .websocket.streams import (agg_trade, blvt_info, blvt_kline,
                                    book_ticker, composite_index,
                                    continuous_kline, diff_book_depth, kline,
                                    liquidation_order, mark_price, mini_ticker,
                                    partial_book_depth, subscription_streams,
                                    ticker, user_data)



import asyncio
import socket

import aiohttp
from loguru import logger


class Ws:

    def __init__(self, **kwargs):
        self.listen_key: str = kwargs.get('listen_key')
        self.reply_timeout: int = kwargs.get('reply_timeout', 180)
        self.ping_timeout: int = kwargs.get('ping_timeout', 300)
        self.sleep_time: int = kwargs.get('sleep_time', 3)

    async def _listen_forever(self, path: str, event: object) -> None:
        _url = self.__create_url(path)
        self.quit = False
        while True:
            if self.quit:
                break
            async with aiohttp.ClientSession() as client:
                async with client.ws_connect(_url, autoclose=False, autoping=False, ssl=True) as self.ws:
                    logger.log('WEBSOCKET', f'Creating new connection {_url}')
                    await self.__dispatch(event)

    async def __dispatch(self, event) -> None:
        while True:
            if self.quit:
                break
            try:
                msg = await self.ws.receive(timeout=self.reply_timeout)
            except (asyncio.TimeoutError, TimeoutError):
                try:
                    await self.ws.ping()
                    msg = await self.ws.receive(timeout=self.ping_timeout)
                    if msg.type == aiohttp.WSMsgType.PONG:
                        continue
                except:
                    logger.log(
                        'WEBSOCKET',
                        f'Ping error - retrying connection in {self.sleep_time} sec (Ctrl-C to quit)')
                    await asyncio.sleep(self.sleep_time)
                    if self.listen_key:
                        self.quit = True
                    break
            except socket.gaierror:
                logger.log(
                    'WEBSOCKET',
                    f'Socket error - retrying connection in {self.sleep_time} sec (Ctrl-C to quit)')
                await asyncio.sleep(self.sleep_time)
                continue
            except ConnectionRefusedError:
                logger.log(
                    'WEBSOCKET',
                    'Nobody seems to listen to this endpoint. Please check the URL.')
                logger.log(
                    'WEBSOCKET',
                    f'Retrying connection in {self.sleep_time} sec (Ctrl-C to quit)')
                await asyncio.sleep(self.sleep_time)
                continue
            else:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    await event(msg.json())
                elif msg.type == aiohttp.WSMsgType.PING:
                    await self.ws.pong()
                elif msg.type == aiohttp.WSMsgType.PONG:
                    logger.log(
                    'WEBSOCKET',
                    'Pong received')
                else:
                    if msg.type == aiohttp.WSMsgType.CLOSE:
                        await self.ws.close()
                    elif msg.type == aiohttp.WSMsgType.ERROR:
                        logger.error(
                            f"Error during receive {self.ws.exception()}")
                    elif msg.type == aiohttp.WSMsgType.CLOSED:
                        pass
                    break

    def __create_url(self, path):
        if self.listen_key:
            _host = 'wss://fstream-auth.binance.com'
            if path.find('listenKey') >= 0:
                return f"{_host}/ws/{self.listen_key}?{path}={self.listen_key}"
            elif path.find('streams') >= 0:
                return f"{_host}/stream?{path}&listenKey={self.listen_key}"
            else:
                return f"{_host}/ws/{path}?listenKey={self.listen_key}"
        else:
            _host = 'wss://fstream.binance.com'
            if path.find('streams') >= 0:
                return f"{_host}/stream?{path}"
            else:
                return f"{_host}/ws/{path}"





async def agg_trade(self, symbol: str, callback_event: object=None):
    """Aggregate Trade Streams
    The Aggregate Trade Streams push market trade information that is aggregated for a single taker order every 100 milliseconds.
    Only market trades will be aggregated, which means the insurance fund trades and ADL trades won't be aggregated.
    
    Stream Name: <symbol>@aggTrade
    https://binance-docs.github.io/apidocs/futures/en/#aggregate-trade-streams
    Update Speed: 100ms
    """
    stream = f"{symbol.lower()}@aggTrade"
    if callback_event is None:
        return stream
    await self._listen_forever(stream, callback_event)

async def mark_price(self, symbol: str, callback_event: object=None):
    """Mark Price Streams
    Mark price and funding rate for all symbols pushed every 3 seconds or every second.
    
    Stream Name: <symbol>@markPrice or <symbol>@markPrice@1s
    https://binance-docs.github.io/apidocs/futures/en/#mark-price-stream
    """
    stream = f"{symbol.lower()}@markPrice@1s"
    if callback_event is None:
        return stream
    await self._listen_forever(stream, callback_event)

async def kline(self, symbol: str, interval: str, callback_event: object=None):
    """Kline/Candlestick Streams
    The Kline/Candlestick Stream push updates to the current klines/candlestick every 250 milliseconds (if existing)
    Stream Name: <symbol>@kline_<interval>
    https://binance-docs.github.io/apidocs/futures/en/#kline-candlestick-streams
    interval:
    m -> minutes; h -> hours; d -> days; w -> weeks; M -> months
    - 1m
    - 3m
    - 5m
    - 15m
    - 30m
    - 1h
    - 2h
    - 4h
    - 6h
    - 8h
    - 12h
    - 1d
    - 3d
    - 1w
    - 1M
    Update Speed: 250ms
    """
    stream = f"{symbol.lower()}@kline_{interval}"
    if callback_event is None:
        return stream
    await self._listen_forever(stream, callback_event)


async def continuous_kline(self, pair: str, contractType: str, interval: str, callback_event: object=None):
    """Continuous Kline/Candlestick Streams
    The Kline/Candlestick Stream push updates to Kline/candlestick bars for a specific contract type. every 250 milliseconds
    Stream Name: <pair>_<contractType>@continuousKline_<interval>
    https://binance-docs.github.io/apidocs/futures/en/#continuous-contract-kline-candlestick-streams
    interval:
    m -> minutes; h -> hours; d -> days; w -> weeks; M -> months
    - 1m
    - 3m
    - 5m
    - 15m
    - 30m
    - 1h
    - 2h
    - 4h
    - 6h
    - 8h
    - 12h
    - 1d
    - 3d
    - 1w
    - 1M
    Update Speed: 250ms
    """
    stream = f"{pair.lower()}_{contractType}@continuousKline_{interval}"
    if callback_event is None:
        return stream
    await self._listen_forever(stream, callback_event)


async def mini_ticker(self, symbol=None, callback_event: object=None):
    """Individual symbol or all symbols mini ticker
    24hr rolling window mini-ticker statistics.
    These are NOT the statistics of the UTC day, but a 24hr rolling window for the previous 24hrs
    Stream Name: <symbol>@miniTicker or
    Stream Name: !miniTicker@arr
    https://binance-docs.github.io/apidocs/futures/en/#individual-symbol-mini-ticker-stream
    https://binance-docs.github.io/apidocs/futures/en/#individual-symbol-ticker-streams
    Update Speed: 500ms for individual symbol, 1000ms for all market symbols
    """
    if symbol is None:
        stream = "!miniTicker@arr"
    else:
        stream = f"{symbol.lower()}@miniTicker"
    if callback_event is None:
        return stream
    await self._listen_forever(stream, callback_event)

async def ticker(self, symbol=None, callback_event: object=None):
    """Individual symbol or all symbols ticker
    24hr rolling window ticker statistics for a single symbol.
    These are NOT the statistics of the UTC day, but a 24hr rolling window from requestTime to 24hrs before.
    Stream Name: <symbol>@ticker or
    Stream Name: !ticker@arr
    https://binance-docs.github.io/apidocs/futures/en/#individual-symbol-ticker-streams
    https://binance-docs.github.io/apidocs/futures/en/#all-market-tickers-streams
    Update Speed: 500ms for individual symbol, 1000ms for all market symbols
    """

    if symbol is None:
        stream = "!ticker@arr"
    else:
        stream = f"{symbol.lower()}@ticker"
    if callback_event is None:
        return stream
    await self._listen_forever(stream, callback_event)

async def book_ticker(self, symbol=None, callback_event: object=None):
    """Individual symbol or all book ticker
    Pushes any update to the best bid or ask's price or quantity in real-time for a specified symbol.
    Stream Name: <symbol>@bookTicker or
    Stream Name: !bookTicker
    https://binance-docs.github.io/apidocs/futures/en/#individual-symbol-book-ticker-streams
    https://binance-docs.github.io/apidocs/futures/en/#all-book-tickers-stream
    
    Update Speed: Real-time
    """
    if symbol is None:
        stream = "!bookTicker"
    else:
        stream = f"{symbol.lower()}@bookTicker"
    if callback_event is None:
        return stream
    await self._listen_forever(stream, callback_event)

async def liquidation_order(self, symbol=None, callback_event: object=None):
    """The Liquidation Order Snapshot Streams push force liquidation order information for specific symbol.
    The All Liquidation Order Snapshot Streams push force liquidation order information for all symbols in the market.
    For each symbol，only the latest one liquidation order within 1000ms will be pushed as the snapshot. If no liquidation happens in the interval of 1000ms, no stream will be pushed.
    
    Stream Name: <symbol>@forceOrder or 
    Stream Name: !forceOrder@arr
    https://binance-docs.github.io/apidocs/futures/en/#liquidation-order-streams
    https://binance-docs.github.io/apidocs/futures/en/#all-market-liquidation-order-streams
    
    Update Speed: 1000ms
    """
    if symbol is None:
        stream = "!forceOrder@arr"
    else:
        stream = f"{symbol.lower()}@forceOrder"
    if callback_event is None:
        return stream
    await self._listen_forever(stream, callback_event)

async def partial_book_depth(self, symbol: str, level: int=5, speed: int=500, callback_event: object=None):
    """Partial Book Depth Streams
    Top bids and asks, Valid are 5, 10, or 20.
    Stream Names: <symbol>@depth<levels> OR <symbol>@depth<levels>@500ms OR <symbol>@depth<levels>@100ms
    https://binance-docs.github.io/apidocs/futures/en/#partial-book-depth-streams
    Update Speed: 250ms, 500ms or 100ms
    """
    stream = f"{symbol.lower()}@depth{level}@{speed}ms"
    if callback_event is None:
        return stream
    await self._listen_forever(stream, callback_event)

async def diff_book_depth(self, symbol: str, speed: int, callback_event: object=None):
    """Diff. Depth Stream
    Order book price and quantity depth updates used to locally manage an order book.
    Stream Name: <symbol>@depth OR <symbol>@depth@500ms OR<symbol>@depth@100ms
    https://binance-docs.github.io/apidocs/futures/en/#diff-book-depth-streams
            
    Update Speed: 250ms, 500ms or 100ms
    """
    stream = f"{symbol.lower()}@depth@{speed}ms"
    if callback_event is None:
        return stream
    await self._listen_forever(stream, callback_event)

async def blvt_info(self, symbol: str, callback_event: object=None):
    """Blvt Info Stream
    get leverage token info
    Stream Name: <tokenName>@tokenNav
    https://binance-docs.github.io/apidocs/futures/en/#blvt-info-streams
    """
    stream = f"{symbol.upper()}@tokenNav"
    if callback_event is None:
        return stream
    await self._listen_forever(stream, callback_event)

async def blvt_kline(self, symbol: str, interval: str, callback_event: object=None):
    """BLVT Kline/Candlestick Streams
    The Kline/Candlestick Stream push updates to the current klines/candlestick every 300 milliseconds (if existing)
    Stream Name: <tokenName>@nav_Kline_<interval>
    https://binance-docs.github.io/apidocs/futures/en/#blvt-nav-kline-candlestick-streams
    interval:
    m -> minutes; h -> hours; d -> days; w -> weeks; M -> months
    - 1m
    - 3m
    - 5m
    - 15m
    - 30m
    - 1h
    - 2h
    - 4h
    - 6h
    - 8h
    - 12h
    - 1d
    - 3d
    - 1w
    - 1M
    Update Speed: 300ms
    """
    stream = f"{symbol.upper()}@nav_Kline_{interval}"
    if callback_event is None:
        return stream
    await self._listen_forever(stream, callback_event)

async def composite_index(self, symbol: str, callback_event: object=None):
    """Composite Index Info Stream
    Composite index information for index symbols pushed every second.
    Stream Name: <symbol>@compositeIndex
    https://binance-docs.github.io/apidocs/futures/en/#composite-index-symbol-information-streams
    
    Update Speed: 1000ms
    """
    stream = f"{symbol.lower()}@compositeIndex"
    if callback_event is None:
        return stream
    await self._listen_forever(stream, callback_event)

async def user_data(self, callback_event: object=None):
    """listen to user data by provided listenkey"""
    stream = "listenKey"
    if callback_event is None:
        return stream
    await self._listen_forever(stream, callback_event)

async def subscription_streams(self, streams: list, callback_event: object):
    stream = 'streams='+'/'.join(streams)
    await self._listen_forever(stream, callback_event)



import sys

import ujson


async def change_position_mode(self, dualSidePosition: str, **kwargs):
    """
    |
    | **Change Position Mode (TRADE)**
    | *Change user's position mode (Hedge Mode or One-way Mode) on EVERY symbol*
    :API endpoint: ``POST /fapi/v1/positionSide/dual``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#change-position-mode-trade
    :parameter dualSidePosition: string

    |
    """
    params = {"dualSidePosition": dualSidePosition, **kwargs}
    url_path = "/fapi/v1/positionSide/dual"
    return await self._fetch(True, 'POST', sys._getframe().f_code.co_name, url_path, **params)


async def get_position_mode(self, kwargs):
    """
    |
    | **Get Current Position Mode (USER_DATA)**
    | *Get user's position mode (Hedge Mode or One-way Mode) on EVERY symbol*
    :API endpoint: ``GET /fapi/v1/positionSide/dual``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#get-current-position-mode-user_data

    |
    """
    url_path = "/fapi/v1/positionSide/dual"
    return await self._fetch(True, 'GET', sys._getframe().f_code.co_name, url_path, kwargs)


async def change_multi_asset_mode(self, multiAssetsMargin: str, **kwargs):
    """
    |
    | **Change Multi-Assets Mode (TRADE)**
    | *Change user's Multi-Assets mode (Multi-Assets Mode or Single-Asset Mode) on Every symbol*
    :API endpoint: ``POST /fapi/v1/multiAssetsMargin``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#change-multi-assets-mode-trade
    :parameter multiAssetsMargin: string; "true": Multi-Assets Mode; "false": Single-Asset Mode

    |
    """
    params = {"multiAssetsMargin": multiAssetsMargin, **kwargs}
    url_path = "/fapi/v1/multiAssetsMargin"
    return await self._fetch(True, 'POST', sys._getframe().f_code.co_name, url_path, **params)


async def get_multi_asset_mode(self, kwargs):
    """
    |
    | **Get Current Multi-Assets Mode (USER_DATA)**
    | *Get user's Multi-Assets mode (Multi-Assets Mode or Single-Asset Mode) on Every symbol*
    :API endpoint: ``GET /fapi/v1/multiAssetsMargin``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#get-current-multi-assets-mode-user_data

    |
    """
    url_path = "/fapi/v1/multiAssetsMargin"
    return await self._fetch(True, 'GET', sys._getframe().f_code.co_name, url_path, kwargs)


async def new_order(self, symbol: str, side: str, type: str, **kwargs):
    """
    |
    | **New Order (TRADE)**
    | *Send a new order*
    :API endpoint: ``POST /fapi/v1/order``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#new-order-trade
    :parameter symbol: string
    :parameter side: string
    :parameter type: string
    :parameter positionSide: optional string. async default BOTH for One-way Mode; LONG or SHORT for Hedge Mode. It must be passed in Hedge Mode.
    :parameter timeInForce: optional string
    :parameter quantity: optional float
    :parameter reduceOnly: optional string
    :parameter price: optional float
    :parameter newClientOrderId: optional string. An unique ID among open orders. Automatically generated if not sent.
    :parameter stopPrice: optional float. Use with STOP/STOP_MARKET or TAKE_PROFIT/TAKE_PROFIT_MARKET orders.
    :parameter closePosition: optional string. true or false; Close-All, use with STOP_MARKET or TAKE_PROFIT_MARKET.
    :parameter activationPrice: optional float. Use with TRAILING_STOP_MARKET orders, async default is the latest price (supporting different workingType).
    :parameter callbackRate: optional float. Use with TRAILING_STOP_MARKET orders, min 0.1, max 5 where 1 for 1%.
    :parameter workingType: optional string. stopPrice triggered by: "MARK_PRICE", "CONTRACT_PRICE". async default "CONTRACT_PRICE".
    :parameter priceProtect: optional string. "TRUE" or "FALSE", async default "FALSE". Use with STOP/STOP_MARKET or TAKE_PROFIT/TAKE_PROFIT_MARKET orders.
    :parameter newOrderRespType: optional float. "ACK" or "RESULT", async default "ACK".

    |
    """
    params = {"symbol": symbol, "side": side, "type": type, **kwargs}
    url_path = "/fapi/v1/order"
    return await self._fetch(True, 'POST', sys._getframe().f_code.co_name, url_path, **params)


async def new_order_test(self, symbol: str, side: str, type: str, **kwargs):
    """
    |
    | **New Test Order (TRADE)**
    | *Send a new test order*
    :API endpoint: ``POST /fapi/v1/order/test``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#new-order-trade
    :parameter symbol: string
    :parameter side: string
    :parameter type: string
    :parameter positionSide: optional string. async default BOTH for One-way Mode; LONG or SHORT for Hedge Mode. It must be passed in Hedge Mode.
    :parameter timeInForce: optional string
    :parameter quantity: optional float
    :parameter reduceOnly: optional string
    :parameter price: optional float
    :parameter newClientOrderId: optional string. An unique ID among open orders. Automatically generated if not sent.
    :parameter stopPrice: optional float. Use with STOP/STOP_MARKET or TAKE_PROFIT/TAKE_PROFIT_MARKET orders.
    :parameter closePosition: optional string. true or false; Close-All, use with STOP_MARKET or TAKE_PROFIT_MARKET.
    :parameter activationPrice: optional float. Use with TRAILING_STOP_MARKET orders, async default is the latest price (supporting different workingType).
    :parameter callbackRate: optional float. Use with TRAILING_STOP_MARKET orders, min 0.1, max 5 where 1 for 1%.
    :parameter workingType: optional string. stopPrice triggered by: "MARK_PRICE", "CONTRACT_PRICE". async default "CONTRACT_PRICE".
    :parameter priceProtect: optional string. "TRUE" or "FALSE", async default "FALSE". Use with STOP/STOP_MARKET or TAKE_PROFIT/TAKE_PROFIT_MARKET orders.
    :parameter newOrderRespType: optional float. "ACK" or "RESULT", async default "ACK".

    |
    """
    params = {"symbol": symbol, "side": side, "type": type, **kwargs}
    url_path = "/fapi/v1/order/test"
    return await self._fetch(True, 'POST', sys._getframe().f_code.co_name, url_path, **params)


async def new_batch_order(self, batch_orders: list):
    """
    |
    | **Place Multiple Orders (TRADE)**
    | *Post a new batch order*
    :API endpoint: ``POST /fapi/v1/batchOrders``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#place-multiple-orders-trade
    :parameter symbol: string
    :parameter side: string
    :parameter type: string
    :parameter positionSide: optional string. async default BOTH for One-way Mode; LONG or SHORT for Hedge Mode. It must be passed in Hedge Mode.
    :parameter timeInForce: optional string
    :parameter quantity: optional float
    :parameter reduceOnly: optional string
    :parameter price: optional float
    :parameter newClientOrderId: optional string. An unique ID among open orders. Automatically generated if not sent.
    :parameter stopPrice: optional float. Use with STOP/STOP_MARKET or TAKE_PROFIT/TAKE_PROFIT_MARKET orders.
    :parameter closePosition: optional string. true or false; Close-All, use with STOP_MARKET or TAKE_PROFIT_MARKET.
    :parameter activationPrice: optional float. Use with TRAILING_STOP_MARKET orders, async default is the latest price (supporting different workingType).
    :parameter callbackRate: optional float. Use with TRAILING_STOP_MARKET orders, min 0.1, max 5 where 1 for 1%.
    :parameter workingType: optional string. stopPrice triggered by: "MARK_PRICE", "CONTRACT_PRICE". async default "CONTRACT_PRICE".
    :parameter priceProtect: optional string. "TRUE" or "FALSE", async default "FALSE". Use with STOP/STOP_MARKET or TAKE_PROFIT/TAKE_PROFIT_MARKET orders.
    :parameter newOrderRespType: optional float. "ACK" or "RESULT", async default "ACK".

    |
    **Notes**
        - Batch orders are processed concurrently, and the order of matching is not guaranteed.
        - The order of returned contents for batch orders is the same as the order of the order list.
            - batchOrders (list): order list. Max 5 orders
        - batchOrders is the list of order parameters in JSON
        - example:
                batch_orders = [
                    {
                        "symbol":"BTCUSDT",
                        "side": "SELL",
                        "type": "LIMIT",
                        "quantity": "0.001",
                        "timeInForce": "GTC",
                        "reduceOnly": "false",
                        "price": "9563.51"
                    },
                    {
                        "symbol":"BTCUSDT",
                        "side": "SELL",
                        "type": "LIMIT",
                        "quantity": "0.001",
                        "timeInForce": "GTC",
                        "reduceOnly": "false",
                        "price": "9613.51"
                    }
                ]
    |
    """
    url_path = "/fapi/v1/batchOrders"
    return await self._fetch(True, 'POST', sys._getframe().f_code.co_name, url_path, batchOrders=ujson.dumps(batch_orders))


async def query_order(self, symbol: str, order_id: int | str):
    """
    |
    | **Query Order (USER_DATA)**
    | *Check an order's status*
    :API endpoint: ``GET /fapi/v1/order``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#query-order-user_data
    :parameter symbol: string
    :parameter order_id: int orderId or string origClientOrderId
    |
    """

    if isinstance(order_id, int):
        params = {"symbol": symbol, "orderId": order_id}
    else:
        params = {"symbol": symbol, "origClientOrderId": order_id}

    url_path = "/fapi/v1/order"
    return await self._fetch(True, 'GET', sys._getframe().f_code.co_name, url_path, **params)


async def cancel_order(self, symbol: str, order_id: int | str):
    """
    |
    | **Cancel Order (TRADE)**
    | *Cancel an active order.*
    :API endpoint: ``DELETE /fapi/v1/order``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#cancel-order-trade
    :parameter symbol: string
    :parameter order_id: int orderId or string origClientOrderId
    |
    """

    if isinstance(order_id, int):
        params = {"symbol": symbol, "orderId": order_id}
    else:
        params = {"symbol": symbol, "origClientOrderId": order_id}

    url_path = "/fapi/v1/order"
    return await self._fetch(True, 'DELETE', sys._getframe().f_code.co_name, url_path, **params)


async def cancel_open_orders(self, symbol: str):
    """
    |
    | **Cancel All Open Orders (TRADE)**
    :API endpoint: ``DELETE /fapi/v1/allOpenOrders``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#cancel-all-open-orders-trade
    :parameter symbol: string
    |
    """
    url_path = "/fapi/v1/allOpenOrders"
    return await self._fetch(True, 'DELETE', sys._getframe().f_code.co_name, url_path, symbol=symbol)


async def cancel_batch_order(self, symbol: str, order_id_list: list[int | str]):
    """
    |
    | **Cancel Multiple Orders (TRADE)**
    | *Cancel a new batch order*
    :API endpoint: ``DELETE /fapi/v1/batchOrders``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#cancel-multiple-orders-trade
    :parameter symbol: string
    :parameter order_id_list: int list; max length 10 e.g. [1234567, 2345678] or string list; max length 10 e.g. ["my_id_1", "my_id_2"], encode the double quotes. No space after comma.
    **Notes**
        - Either orderIdList or origClientOrderIdList must be sent.
    |
    """

    url_path = "/fapi/v1/batchOrders"
    if isinstance(order_id_list[0], int):
        params = {"symbol": symbol, "orderIdList": ujson.dumps(order_id_list)}
    else:
        params = {"symbol": symbol,
                  "origClientOrderIdList": ujson.dumps(order_id_list)}

    return await self._fetch(True, 'DELETE', sys._getframe().f_code.co_name, url_path, **params)


async def countdown_cancel_order(self, symbol: str, countdownTime: int):
    """
    |
    | **Auto-Cancel All Open Orders (TRADE)**
    | *Cancel all open orders of the specified symbol at the end of the specified countdown.*
    :API endpoint: ``POST /fapi/v1/countdownCancelAll``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#auto-cancel-all-open-orders-trade
    :parameter symbol: string
    :parameter countdownTime: int list; countdown time, 1000 for 1 second. 0 to cancel the timer.
    **Notes**
        - The endpoint should be called repeatedly as heartbeats so that the existing countdown time can be canceled and replaced by a new one.
        - Example usage:
            - Call this endpoint at 30s intervals with an countdownTime of 120000 (120s).
            - If this endpoint is not called within 120 seconds, all your orders of the specified symbol will be automatically canceled.
            - If this endpoint is called with an countdownTime of 0, the countdown timer will be stopped.
        - The system will check all countdowns approximately every 10 milliseconds, so please note that sufficient redundancy should be considered when using this function. 
        - We do not recommend setting the countdown time to be too precise or too small.
    """
    url_path = "/fapi/v1/countdownCancelAll"
    return await self._fetch(True, 'POST', sys._getframe().f_code.co_name, url_path, symbol=symbol, countdownTime=countdownTime)


async def get_open_orders(self, symbol: str, order_id: int | str):
    """
    |
    | **Query Current Open Order (USER_DATA)**
    | *Get all open orders on a symbol.*
    :API endpoint: ``GET /fapi/v1/openOrder``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#query-current-open-order-user_data
    :parameter symbol: string
    :parameter orderId: optional int
    :parameter origClientOrderId: optional int

    **Notes**
        - Either orderId or origClientOrderId must be sent
        - If the queried order has been filled or cancelled, the error message "Order does not exist" will be returned.
    """

    url_path = "/fapi/v1/openOrder"
    if isinstance(order_id, int):
        params = {"symbol": symbol, "orderId": order_id}
    else:
        params = {"symbol": symbol, "origClientOrderId": order_id}

    return await self._fetch(True, 'GET', sys._getframe().f_code.co_name, url_path, **params)


async def get_orders(self, symbol: str = None):
    """
    |
    | **Current All Open Orders (USER_DATA)**
    | *Get all open orders on a symbol. Careful when accessing this with no symbol.*
    | *If the symbol is not sent, orders for all symbols will be returned in an array.*
    :API endpoint: ``GET /fapi/v1/openOrders``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#current-all-open-orders-user_data
    :parameter symbol: optional string
    |
    """
    url_path = "/fapi/v1/openOrders"
    return await self._fetch(True, 'GET', sys._getframe().f_code.co_name, url_path, symbol=symbol)


async def get_all_orders(self, symbol: str, **kwargs):
    """
    |
    | **All Orders (USER_DATA)**
    | *Get all account orders; active, canceled, or filled.*
    :API endpoint: ``GET /fapi/v1/allOrders``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#all-orders-user_data
    :parameter symbol: string
    :parameter orderId: optional int
    :parameter startTime: optional int
    :parameter endTime: optional int
    :parameter limit: optional int
    |
    """
    url_path = "/fapi/v1/allOrders"
    params = {"symbol": symbol, **kwargs}
    return await self._fetch(True, 'GET', sys._getframe().f_code.co_name, url_path, **params)


async def balance(self):
    """
    |
    | **Futures Account Balance V2 (USER_DATA)**
    | *Get current account balance*
    :API endpoint: ``GET /fapi/v2/balance``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#futures-account-balance-v2-user_data
    |
    """

    url_path = "/fapi/v2/balance"
    return await self._fetch(True, 'GET', sys._getframe().f_code.co_name, url_path)


async def account(self):
    """
    |
    | **Account Information V2 (USER_DATA)**
    | *Get current account information*
    :API endpoint: ``GET /fapi/v2/account``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#account-information-v2-user_data

    |
    """

    url_path = "/fapi/v2/account"
    return await self._fetch(True, 'GET', sys._getframe().f_code.co_name, url_path)


async def change_leverage(self, symbol: str, leverage: int):
    """
    |
    | **Change Initial Leverage (TRADE)**
    | *Change user's initial leverage of specific symbol market.*
    :API endpoint: ``POST /fapi/v1/leverage``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#change-initial-leverage-trade
    :parameter symbol: string
    :parameter leverage: int; target initial leverage: int from 1 to 125.
    |
    """
    url_path = "/fapi/v1/leverage"
    return await self._fetch(True, 'POST', sys._getframe().f_code.co_name, url_path, symbol=symbol, leverage=leverage)


async def change_margin_type(self, symbol: str, marginType: str):
    """
    |
    | **Change margin type (TRADE)**
    | *Change user's margin type of specific symbol market.*
    :API endpoint: ``POST /fapi/v1/marginType``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#change-margin-type-trade
    :parameter symbol: string
    :parameter marginType: string; ISOLATED, CROSSED.
    |
    """
    url_path = "/fapi/v1/marginType"
    return await self._fetch(True, 'POST', sys._getframe().f_code.co_name, url_path, symbol=symbol, marginType=marginType)


async def modify_isolated_position_margin(self, symbol: str, amount: float, type: int, **kwargs):
    """
    |
    | **Modify Isolated Position Margin (TRADE)**
    :API endpoint: ``POST /fapi/v1/positionMargin``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#modify-isolated-position-margin-trade
    :parameter symbol: string
    :parameter amount: float
    :parameter type: int; 1: Add position margin, 2: Reduce position margin
    :parameter positionSide: optional string; async default BOTH for One-way Mode, LONG or SHORT for Hedge Mode. It must be sent with Hedge Mode.
    |
    """
    url_path = "/fapi/v1/positionMargin"
    params = {"symbol": symbol, "amount": amount, "type": type, **kwargs}
    return await self._fetch(True, 'POST', sys._getframe().f_code.co_name, url_path, **params)


async def get_position_margin_history(self, symbol: str, **kwargs):
    """
    |
    | **Get Position Margin Change History (TRADE)**
    | *Get position margin history on a symbol.*
    :API endpoint: ``GET /fapi/v1/positionMargin/history``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#get-position-margin-change-history-trade
    :parameter symbol: string
    :parameter type: optional int; 1: Add position margin, 2: Reduce position margin.
    :parameter startTime: optional int
    :parameter endTime: optional int
    :parameter limit: optional int; async default: 500.

    |
    """
    url_path = "/fapi/v1/positionMargin/history"
    params = {"symbol": symbol, **kwargs}

    return await self._fetch(True, 'GET', sys._getframe().f_code.co_name, url_path, **params)


async def get_position_risk(self, symbol: str = None):
    """
    |
    | **Position Information V2 (USER_DATA)**
    | *Get current position information.*
    :API endpoint: ``GET /fapi/v2/positionRisk``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#position-information-v2-user_data
    :parameter symbol: string

    |
    """

    url_path = "/fapi/v2/positionRisk"

    return await self._fetch(True, 'GET', sys._getframe().f_code.co_name, url_path, symbol=symbol)


async def get_account_trades(self, symbol: str, **kwargs):
    """
    |
    | **Account Trade List (USER_DATA)**
    | *Get trades for a specific account and symbol.*
    :API endpoint: ``GET /fapi/v1/userTrades``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#account-trade-list-user_data
    :parameter symbol: string
    :parameter startTime: optional int
    :parameter endTime: optional int
    :parameter fromId: optional int; trade ID to fetch from, async default gets most recent trades.
    :parameter limit: optional int; async default: 500, max: 1000.

    **Notes**
        - If startTime and endTime are both not sent, then the last 7 days' data will be returned.
        - The time between startTime and endTime cannot be longer than 7 days.
        - The parameter fromId cannot be sent with startTime or endTime.
    |
    """

    url_path = "/fapi/v1/userTrades"
    params = {"symbol": symbol, **kwargs}

    return await self._fetch(True, 'GET', sys._getframe().f_code.co_name, url_path, **params)


async def get_income_history(self, kwargs):
    """
    |
    | **Get Income History (USER_DATA)**
    | *Get trades for a specific account and symbol.*
    :API endpoint: ``GET /fapi/v1/income``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#get-income-history-user_data
    :parameter symbol: optional string
    :parameter incomeType: optional string; "TRANSFER", "WELCOME_BONUS", "REALIZED_PNL", "FUNDING_FEE", "COMMISSION" and "INSURANCE_CLEAR".
    :parameter startTime: optional int; timestamp in ms to get funding from INCLUSIVE.
    :parameter endTime: optional int; timestamp in ms to get funding from INCLUSIVE.
    :parameter limit: optional int; async default: 100, max: 1000.

    **Notes**
        - If neither startTime nor endTime is sent, the recent 7-day data will be returned.
        - If incomeType is not sent, all kinds of flow will be returned
        - "trandId" is unique in the same incomeType for a user
    """

    url_path = "/fapi/v1/income"
    return await self._fetch(True, 'GET', sys._getframe().f_code.co_name, url_path, kwargs)


async def leverage_brackets(self, kwargs):
    """
    |
    | **Notional and Leverage Brackets (USER_DATA)**
    | *Get notional and leverage bracket.*
    :API endpoint: ``GET /fapi/v1/leverageBracket``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#notional-and-leverage-brackets-user_data
    :parameter symbol: optional string

    |
    """

    url_path = "/fapi/v1/leverageBracket"
    return await self._fetch(True, 'GET', sys._getframe().f_code.co_name, url_path, kwargs)


async def adl_quantile(self, kwargs):
    """
    |
    | **Position ADL Quantile Estimation (USER_DATA)**
    | *Get Position ADL Quantile Estimation*
    :API endpoint: ``GET /fapi/v1/adlQuantile``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#position-adl-quantile-estimation-user_data
    :parameter symbol: optional string

    **Notes**
        - Values update every 30s.
        - Values 0, 1, 2, 3, 4 shows the queue position and possibility of ADL from low to high.
        - For positions of the symbol are in One-way Mode or isolated margined in Hedge Mode, "LONG", "SHORT", and "BOTH" will be returned to show the positions' adl quantiles of different position sides.
        - If the positions of the symbol are crossed margined in Hedge Mode:
            - "HEDGE" as a sign will be returned instead of "BOTH"
        - A same value caculated on unrealized pnls on long and short sides' positions will be shown for "LONG" and "SHORT" when there are positions in both of long and short sides.
    |
    """

    url_path = "/fapi/v1/adlQuantile"
    return await self._fetch(True, 'GET', sys._getframe().f_code.co_name, url_path, kwargs)


async def force_orders(self, kwargs):
    """
    |
    | **User's Force Orders (USER_DATA)**
    | *Get User's Force Orders*
    :API endpoint: ``GET /fapi/v1/forceOrders``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#user-39-s-force-orders-user_data
    :parameter symbol: optional string
    :parameter autoCloseType: optional string; "LIQUIDATION" for liquidation orders, "ADL" for ADL orders.
    :parameter startTime: optional int
    :parameter endTime: optional int
    :parameter Limit: optional int; async default 50, max 100.

    **Notes**
        - If "autoCloseType" is not sent, orders with both of the types will be returned
        - If "startTime" is not sent, data within 7 days before "endTime" can be queried
    |
    """

    url_path = "/fapi/v1/forceOrders"

    return await self._fetch(True, 'GET', sys._getframe().f_code.co_name, url_path, kwargs)


async def api_trading_status(self, kwargs):
    """
    |
    | **User API Trading Quantitative Rules Indicators (USER_DATA)**
    | *Get User API Trading Quantitative Rules Indicators*
    :API endpoint: ``GET /fapi/v1/apiTradingStatus``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#user-api-trading-quantitative-rules-indicators-user_data
    :parameter symbol: optional string

    |
    """

    url_path = "/fapi/v1/apiTradingStatus"

    return await self._fetch(True, 'GET', sys._getframe().f_code.co_name, url_path, kwargs)


async def commission_rate(self, symbol: str):
    """
    |
    | **User Commission Rate (USER_DATA)**
    | *Get commission rate of symbol*
    :API endpoint: ``GET /fapi/v1/commissionRate``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#user-commission-rate-user_data
    :parameter symbol: string

    |
    """
    url_path = "/fapi/v1/commissionRate"
    return await self._fetch(True, 'GET', sys._getframe().f_code.co_name, url_path, symbol=symbol)



import asyncio
import hmac
from copy import deepcopy
from hashlib import sha256
from time import sleep, time
from typing import Dict
from urllib.parse import urlencode

import ujson
from aiohttp import ClientSession
from loguru import logger

from .error import BinanceApiException


class Api():

    shift_seconds = 0
    recconect = 0
    weight = 0
    timeout = 5

    def __init__(self, **kwargs):
        self.key = kwargs.get('key')
        self.secret = kwargs.get('secret')
        self.show_limit_usage = kwargs.get('show_limit_usage')
        self.show_header = kwargs.get('show_header')
        self.host = 'https://testnet.binancefuture.com' \
            if kwargs.get('testnet') \
            else 'https://fapi.binance.com'
        self.session: ClientSession = None
        self.headers = {}
        self.agent = kwargs.get('agent', 'aio-binance-futures')
        self.version = kwargs.get('version')

    def __set_shift_seconds(self, seconds):
        Api.shift_seconds = seconds

    def __check_response(self, json_wrapper: Dict):
        code = 200
        msg = ""
        if isinstance(json_wrapper, list):
            for item in json_wrapper:
                if isinstance(item, Dict) and \
                    item.get('code') and \
                        item['code'] != 200:
                    code = item['code']
                    msg = json_wrapper
                    break
        else:
            if json_wrapper.get("code"):
                code = int(json_wrapper["code"])
                msg = json_wrapper.get("msg", "")
        if code != 200:
            if code == -1021:
                self.__set_shift_seconds(Api.shift_seconds - 1)
            elif code == -1003:
                start = msg.find(' ', msg.find('until'))
                end = msg.find('.', start)
                timer = int((int(msg[start:end-1]) - int(time() * 100)) / 100)
                logger.log('API', f"Binance banned IP. I'll be waiting {timer} sec.")
                sleep(timer)
            else:
                raise BinanceApiException(code, msg)

    def __crypto_key(self, params: Dict) -> None:
        params.update({
            'recvWindow': 60000,
            'timestamp': int((time() + Api.shift_seconds - 1) * 1000)
        })
        params_str = urlencode(params).encode('utf-8')
        sign = hmac.new(
            key=bytearray(self.secret, encoding='utf-8'),
            msg=params_str,
            digestmod=sha256
        ).hexdigest()
        params.update({'signature': str(sign)})

    async def __recconect(self, args, kwargs, sleeping, err):
        if Api.recconect > 10:
            raise BinanceApiException(
                BinanceApiException.RUNTIME_ERROR, f"(Binance Futures Api) [Connector Error]: Exception {err}")
        else:
            logger.error(err)
            Api.timeout += 2
            Api.recconect += 1
            await asyncio.sleep(sleeping)
            return await self.fetch(*args, **kwargs)

    async def _fetch(self, *args, **kwargs) -> Dict:
        _kwargs_deep = deepcopy(kwargs)
        self.headers = {
            'Content-Type': 'application/json',
            'user-agent': self.agent,
            "client_SDK_Version": f"aio-binance-futures {self.version}"
        }
        if args[0]:
            assert self.key is not None, \
                "For job this function needs api key binance, please add in Client()"
            assert self.secret is not None, \
                "For job this function needs api secret binance, please add in Client()"
            self.headers.update({"X-MBX-APIKEY": self.key})
            self.__crypto_key(kwargs)
        url = self.host + args[3]
        request_data = {
            'method': args[1],
            'url': url,
            'timeout': Api.timeout
        }
        if args[1] == 'GET':
            request_data['headers'] = self.headers
            request_data['params'] = kwargs if kwargs.keys() else None
        else:
            self.headers.update({
                'Content-Type': 'application/x-www-form-urlencoded'
            })
            request_data['headers'] = self.headers
            request_data['data'] = kwargs if kwargs.keys() else None
        try:
            start_time = time() * 1000
            if self.session:
                async with self.session.request(**request_data) as responce:
                    _responce = await responce.text()
            else:
                async with ClientSession() as session:
                    async with session.request(**request_data) as responce:
                        _responce = await responce.text()
            Api.weight = responce.headers['X-MBX-USED-WEIGHT-1M']
        except Exception as err:
            return await self.__recconect(args, _kwargs_deep, 3, err)
        else:
            if Api.recconect > 0:
                Api.recconect = 0
                Api.timeout = 2
            logger.log(
                'API',
                "      Function {}() work Ok! Ping: {:.2f} ms. Weight: {}".format(
                    args[2],
                    (time() * 1000)-start_time,
                    Api.weight))
            try:
                res_json = ujson.loads(_responce)
            except ValueError:
                raise BinanceApiException(
                    BinanceApiException.INPUT_ERROR,
                    f"(Binance Futures Api) [ValueError]: {_responce}")
            else:
                self.__check_response(res_json)
                if self.show_limit_usage:
                    res_json['limit_usage'] = Api.weight
                if self.show_header:
                    res_json['header'] = responce.headers
                return res_json



import sys

async def new_listen_key(self):
    """
    |
    | **Create a ListenKey (USER_STREAM)**
    :API endpoint: ``POST /fapi/v1/listenKey``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#start-user-data-stream-user_stream
    |
    """

    url_path = "/fapi/v1/listenKey"
    return await self._fetch(True, 'POST', sys._getframe().f_code.co_name, url_path)


async def update_listen_key(self):
    """
    |
    | **Ping/Keep-alive a ListenKey (USER_STREAM)**
    :API endpoint: ``PUT /fapi/v1/listenKey``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#keepalive-user-data-stream-user_stream
    :parameter listenKey: string
    |
    """
    url_path = "/fapi/v1/listenKey"
    return await self._fetch(True, 'PUT', sys._getframe().f_code.co_name, url_path)


async def close_listen_key(self):
    """
    |
    | **Close a ListenKey (USER_STREAM)**
    :API endpoint: ``DELETE /fapi/v1/listenKey``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#close-user-data-stream-user_stream
    :parameter listenKey: string
    |
    """
    url_path = "/fapi/v1/listenKey"
    return await self._fetch(True, 'DELETE', sys._getframe().f_code.co_name, url_path)



import sys

async def ping(self):
    """
    |
    | **Test Connectivity**
    | *Test connectivity to the Rest API.*
    :API endpoint: ``GET /fapi/v1/ping``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#test-connectivity
    |
    """
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/ping")


async def time(self):
    """
    |
    | **Check Server Time**
    | *Test connectivity to the Rest API and get the current server time.*
    :API endpoint: ``GET /fapi/v1/time``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#check-server-time
    |
    """
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/time")


async def exchange_info(self) -> dict:
    """
    **Exchange Information**\n
    *Current exchange trading rules and symbol information.*\n
    :API endpoint: ``GET /fapi/v1/exchangeInfo``\n
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#exchange-information\n

    Returns:
        dict: Exchange Information
    """
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/exchangeInfo")


async def depth(self, symbol: str, **kwargs):
    """
    |
    | **Get Orderbook**
    :API endpoint: ``GET /fapi/v1/depth``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#order-book
    :parameter symbol: string; the trading symbol.
    :parameter limit: optional int; limit the results. async default 500, valid limits: [5, 10, 20, 50, 100, 500, 1000].
    |
    """
    params = {"symbol": symbol, **kwargs}
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/depth", **params)


async def trades(self, symbol: str, **kwargs):
    """
    |
    | **Get Recent Market Trades**
    :API endpoint: ``GET /fapi/v1/trades``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#recent-trades-list
    :parameter symbol: string; the trading symbol.
    :parameter limit: optional int; limit the results. async default 500, max 1000.
    |
    """
    params = {"symbol": symbol, **kwargs}
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/trades", **params)


async def historical_trades(self, symbol: str, **kwargs):
    """
    |
    | **Old Trade Lookup**
    | *Get older market historical trades.*
    :API endpoint: ``GET /fapi/v1/historicalTrades``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#old-trades-lookup-market_data
    :parameter symbol: string; the trading symbol.
    :parameter limit: optional int; limit the results. async default 500, max 1000.
    :parameter formId: optional int; trade ID to fetch from. async default gets most recent trades.
    |
    """
    params = {"symbol": symbol, **kwargs}
    return self.limit_request("GET", "/fapi/v1/historicalTrades", **params)


async def agg_trades(self, symbol: str, **kwargs):
    """
    |
    | **Compressed/Aggregate Trades List**
    | *Get compressed, aggregate market trades. Market trades that fill at the time, from the same order, with the same price will have the quantity aggregated.*
    
    :API endpoint: ``GET /fapi/v1/aggTrades``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#compressed-aggregate-trades-list
    :parameter symbol: string; the trading symbol.
    :parameter limit: optional int; limit the results. async default 500, max 1000.
    :parameter formId: optional int; ID to get aggregate trades from INCLUSIVE.
    :parameter startTime: optional int; timestamp in ms to get aggregate trades from INCLUSIVE.
    :parameter endTime: optional int; timestamp in ms to get aggregate trades from INCLUSIVE.
    |
    """
    params = {"symbol": symbol, **kwargs}
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/aggTrades", **params)


async def klines(self, symbol: str, interval: str, **kwargs):
    """
    |
    | **Kline/Candlestick Data**
    | *Kline/candlestick bars for a symbol. Klines are uniquely identified by their open time.*
    :API endpoint: ``GET /fapi/v1/klines``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#kline-candlestick-data
    :parameter symbol: string; the trading symbol.
    :parameter interval: string; the interval of kline, e.g 1m, 5m, 1h, 1d, etc. (see more in https://binance-docs.github.io/apidocs/futures/en/#public-endpoints-info)
    :parameter limit: optional int; limit the results. async default 500, max 1000.
    :parameter startTime: optional int
    :parameter endTime: optional int
    |
    """
    params = {"symbol": symbol, "interval": interval, **kwargs}
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/klines", **params)


async def continuous_klines(self, pair: str, contractType: str, interval: str, **kwargs):
    """
    |
    | **Continuous Kline/Candlestick Data**
    | *Kline/candlestick bars for a specific contract type. Klines are uniquely identified by their open time.*  
    
    :API endpoint: ``GET /fapi/v1/continuousKlines``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#continuous-contract-kline-candlestick-data
    :parameter pair: string; the trading pair.
    :parameter contractType: string; PERPETUAL, CURRENT_MONTH, NEXT_MONTH, CURRENT_QUARTER, NEXT_QUARTER.
    :parameter interval: string; the interval of kline, e.g 1m, 5m, 1h, 1d, etc. (see more in https://binance-docs.github.io/apidocs/futures/en/#public-endpoints-info)
    :parameter limit: optional int; limit the results. async default 500, max 1000.
    :parameter startTime: optional int
    :parameter endTime: optional int
    |
    """
    params = {"pair": pair, "contractType":contractType, "interval": interval, **kwargs}
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/continuousKlines", **params)


async def index_price_klines(self, pair: str, interval: str, **kwargs):
    """
    |
    | **Kline/Candlestick Data for the index price of a pair.**
    | *Klines are uniquely identified by their open time.*   
    
    :API endpoint: ``GET /fapi/v1/indexPriceKlines``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#index-price-kline-candlestick-data
    :parameter pair: string; the trading pair.
    :parameter interval: string; the interval of kline, e.g 1m, 5m, 1h, 1d, etc. (see more in https://binance-docs.github.io/apidocs/futures/en/#public-endpoints-info)
    :parameter limit: optional int; limit the results. async default 500, max 1000.
    :parameter startTime: optional int
    :parameter endTime: optional int
    |
    """
    params = {"pair": pair, "interval": interval, **kwargs}
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/indexPriceKlines", **params)


async def mark_price_klines(self, symbol: str, interval: str, **kwargs) -> list:
    """
    |
    | **Kline/candlestick bars for the mark price of a symbol.**\n
    | *Klines are uniquely identified by their open time.*\n
    
    :API endpoint: ``GET /fapi/v1/markPriceKlines``\n
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#mark-price-kline-candlestick-data\n

    Args:
        symbol (str): the trading symbol.\n
        interval (str): the interval of kline, e.g 1m, 5m, 1h, 1d, etc. (see more in https://binance-docs.github.io/apidocs/futures/en/#public-endpoints-info)\n
        limit optional(int): limit the results. async default 500, max 1000.\n
        startTime optional(int)\n
        endTime: optional(int)\n

    Returns:
        [list]: Kline/candlestick bars List
    """
    params = {"symbol": symbol, "interval": interval, **kwargs}
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/markPriceKlines", **params)


async def mark_price(self, symbol: str):
    """
    |
    | **Mark Price and Funding Rate**
    :API endpoint: ``GET /fapi/v1/premiumIndex``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#mark-price
    :parameter symbol: string; the trading symbol.
    |
    """
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/premiumIndex", symbol=symbol)


async def funding_rate(self, symbol: str,  **kwargs):
    """
    |
    | **Funding Rate History
    :API endpoint: ``GET /fapi/v1/fundingRate``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#get-funding-rate-history
    :parameter symbol: string; the trading symbol.
    :parameter limit: optional int; limit the results. async default 500, max 1000.
    :parameter startTime: optional int
    :parameter endTime: optional int
    **Notes**
        - If startTime and endTime are not sent, the most recent limit datas are returned.
        - If the number of data between startTime and endTime is larger than limit, return as startTime + limit.
        - In ascending order.
    |
    """
    params = {"symbol": symbol, **kwargs}
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/fundingRate", **params)


async def ticker_24hr_price_change(self, symbol: str = None):
    """
    |
    | **24 hour rolling window price change statistics.**
    | *Careful when accessing this with no symbol.*
    | *If the symbol is not sent, tickers for all symbols will be returned in an array.*
    :API endpoint: ``GET /fapi/v1/ticker/24hr``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#24hr-ticker-price-change-statistics
    :parameter symbol: string; the trading symbol.
    |
    """
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/ticker/24hr", symbol=symbol)


async def ticker_price(self, symbol: str = None):
    """
    |
    | **Latest price for a symbol or symbols.**
    :API endpoint: ``GET /fapi/v1/ticker/price``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#symbol-price-ticker
    :parameter symbol: optional string; the trading symbol.
    **Notes**
        - If the symbol is not sent, prices for all symbols will be returned in an array.
    |
    """
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/ticker/price", symbol=symbol)


async def book_ticker(self, symbol: str = None):
    """
    |
    | **Best price/qty on the order book for a symbol or symbols.**
    :API endpoint: ``GET /fapi/v1/ticker/bookTicker``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#symbol-order-book-ticker
    :parameter symbol: optional string; the trading symbol.
    
    **Notes**
        - If the symbol is not sent, bookTickers for all symbols will be returned in an array.
    |
    """
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/ticker/bookTicker", symbol=symbol)


async def open_interest(self, symbol: str):
    """
    |
    | **Get present open interest of a specific symbol.**
    :API endpoint: ``GET /fapi/v1/openInterest``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#open-interest
    :parameter symbol: string; the trading symbol.
    |
    """
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/ticker/bookTicker", symbol=symbol)


async def blvt_kline(self, symbol: str, interval: str, **kwargs):
    """
    |
    | **Get Historical BLVT NAV Kline**
    
    :API endpoint: ``GET /fapi/v1/lvtKlines``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#historical-blvt-nav-kline-candlestick
    :parameter symbol: string; the trading symbol.
    :parameter period: string; the period of open interest, "5m", "15m", "30m", "1h", "2h", "4h", "6h", "12h", "1d". (see more in https://binance-docs.github.io/apidocs/futures/en/#public-endpoints-info)
    :parameter limit: optional int; limit the results. async default 500, max 1000.
    :parameter startTime: optional int
    :parameter endTime: optional int
    **Notes**
        - If startTime and endTime are not sent, the most recent data is returned.
    |
    """
    params = {"symbol": symbol, "interval": interval, **kwargs}
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/lvtKlines", **params)


async def index_info(self, symbol: str = None):
    """
    |
    | **Get Index Composite**
    :API endpoint: ``GET /fapi/v1/indexInfo``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#composite-index-symbol-information
    :parameter symbol: optional string; the trading symbol.
    **Notes**
        - Only for composite index symbols.
    |
    """
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/indexInfo", symbol=symbol)


async def asset_Index(self, symbol: str = None):
    """
    |
    | **Get asset index for Multi-Assets mode**
    :API endpoint: ``GET /fapi/v1/assetIndex``
    :API doc: https://binance-docs.github.io/apidocs/futures/en/#multi-assets-mode-asset-index
    :parameter symbol: optional string; Asset pair in multi asset mode (ex: BTCUSD).
    |
    """
    return await self._fetch(False, 'GET', sys._getframe().f_code.co_name, "/fapi/v1/assetIndex", symbol=symbol)